/*wbqrs(length transformation based algorithm -wqrs + bp-ECG characterisric point matching method
  this algorithm is developed based on wqrs source code, and source code is released on physionet web site*/
#include <stdio.h>
#include <math.h>
#include <wfdb/wfdb.h>
#include <wfdb/ecgcodes.h>
#include <wfdb/ecgmap.h>
#include <gsl/gsl_wavelet.h>

#define BUFLN  16384	/* must be a power of 2, see ltsamp() */
#define EYE_CLS 0.25    /* eye-closing period is set to 0.25 sec (250 ms) */ 
#define MaxQRSw 0.13    /* maximum QRS width (130ms) */                        
#define NDP	 2.5   	/* adjust threshold if no QRS found in NDP seconds */
#define PWFreqDEF 60    /* power line (mains) frequency, in Hz (default) */
#define TmDEF	 100	/* minimum threshold value (default) */

char *pname;		/* the name by which this program was invoked */
double lfsc;		/* length function scale constant */
int *ebuf;
int nsig;		/* number of input signals */
int LPn, LP2n;          /* filter parameters (dependent on sampling rate) */
int LTwindow;           /* LT window size */
int PWFreq = PWFreqDEF;	/* power line (mains) frequency, in Hz */
int sig = -1;	        /* signal number of signal to be analyzed */
int Tm = TmDEF;		/* minimum threshold value */
WFDB_Sample *lbuf = NULL;

/* ltsamp() returns a sample of the length transform of the input at time t.
   Since this program analyzes only one signal, ltsamp() does not have an
   input argument for specifying a signal number; rather, it always filters
   and returns samples from the signal designated by the global variable
   'sig'.  The caller must never "rewind" by more than BUFLN samples (the
   length of ltsamp()'s buffers). */

WFDB_Sample ltsamp(WFDB_Time t)
{
    int dy;
    static int Yn, Yn1, Yn2;
    static WFDB_Time tt = (WFDB_Time)-1L;

    if (lbuf == NULL) {
	lbuf = (WFDB_Sample *)malloc((unsigned)BUFLN*sizeof(WFDB_Sample));
	ebuf = (int *)malloc((unsigned)BUFLN * sizeof(int));
	if (lbuf && ebuf) {
	    for (ebuf[0] = sqrt(lfsc), tt = 1L; tt < BUFLN; tt++)
		ebuf[tt] = ebuf[0];
	    if (t > BUFLN) tt = (WFDB_Time)(t - BUFLN);
	    else tt = (WFDB_Time)-1L;
	    Yn = Yn1 = Yn2 = 0;
	}
	else {
	    (void)fprintf(stderr, "%s: insufficient memory\n", pname);
	    exit(2);
	}
    }
    if (t < tt - BUFLN) {
	fprintf(stderr, "%s: ltsamp buffer too short\n", pname);
	exit(2);
    }
    while (t > tt) {
	static int aet = 0, et;
	WFDB_Sample v0, v1, v2;

	Yn2 = Yn1;
	Yn1 = Yn;
	if ((v0 = sample(sig, tt)) != WFDB_INVALID_SAMPLE &&
		(v1 = sample(sig, tt-LPn)) != WFDB_INVALID_SAMPLE &&
		(v2 = sample(sig, tt-LP2n)) != WFDB_INVALID_SAMPLE)
	    Yn = 2*Yn1 - Yn2 + v0 - 2*v1 + v2;
	dy = (Yn - Yn1) / LP2n;		/* lowpass derivative of input */
	et = ebuf[(++tt)&(BUFLN-1)] = sqrt(lfsc +dy*dy); /* length transform */
	lbuf[(tt)&(BUFLN-1)] = aet += et - ebuf[(tt-LTwindow)&(BUFLN-1)];
	/* lbuf contains the average of the length-transformed samples over
	   the interval from tt-LTwindow+1 to tt */
    }
    return (lbuf[t&(BUFLN-1)]);
}

main(int argc, char **argv)
{
    /* Variables for WQRS */ 
    char *p;
    char *record = NULL;		/* input record name */
    float sps;			     	/* sampling frequency, in Hz (SR) */
    float samplingInterval;		/* sampling interval, in milliseconds */
    int i, max, min, minutes = 0, onset, timer, vflag = 0;
    int dflag = 0;		     	/* if non-zero, dump raw and filtered samples only;  do not run detector */
    int jflag = 0;		     	/* if non-zero, annotate J-points */
    int Rflag = 0;		     	/* if non-zero, resample at 120 or 150 Hz */
    int EyeClosing;			/* eye-closing period, related to SR */
    int ExpectPeriod;			/* if no QRS is detected over this period, 
					    the threshold is automatically reduced
					    to a minimum value;  the threshold is
					    restored upon a detection */
    double Ta, T0;		     	/* high and low detection thresholds */
    WFDB_Anninfo a;
    WFDB_Annotation annot;
    WFDB_Gain gain;
    WFDB_Time from = 0L, next_minute, spm, t, tj, to = 0L, tt, t1;
    static int gvmode = WFDB_GVPAD | WFDB_LOWRES;
    char *prog_name();
    void help();

    static WFDB_Siginfo *s;
    int *v;
    int tpq;

    /* Variables for Score-based QRS detector */
    int lengthWindow = 2^10;		/* Window size: 2^10 @ 250Hz, 2^12 @ 360Hz  */

    gsl_wavelet *w = gsl_wavelet_alloc(gsl_wavelet_daubechies, 8);
    gsl_wavelet_workspace *work;

    /* Variables for BP-ECG Characteristic point matching */
    int reset_time = 0;			/* reset global timer */
    int current_time = 0;		/* location of current processing */
    int base_time = 0;			/* location of global timer */
    int qrs_interval = 0;		/* estimated interval between BP and ECG signals */
    int qrs_time = 0;			/* location of detected qrs complex using BP-ECG intervals */
    int offset_time = 80;		/* Offset for qrs interval */
    int bp_array[150]={0,};		/* array for BP signal */
    int ecg_array[150]={0,};		/* array for ECG signal */
    int bp_t[50]={0,};			/* temporal buffer for sliding window on BP signal */
    int ecg_t[50]={0,};			/* temporal buffer for sliding window on ECG signal */
    int bp_temp[250]={0,};		/* buffer for BP signal */
    int ecg_temp[250]={0,};		/* buffer for ECG signal */

					/* We devide the searching window on BP signal into three sections */
    int bp_left = 0;			/* index of maximum point in left section */
    int bp_mid = 0;			/* index of maximum point in middle section */
    int bp_right = 0;			/* index of maximum point in right section */
    int bp_max=0;			/* index of maximum point in searching window on BP signal */
    int ecg_max=0;			/* index of maximum point in searching window on ECG signal */

    int cntBpMax = 0;			/* count the number of observation */
    int numObservation = 50;		/* number of observation for initial intervals between BP-ECG signals */
    int qrsDetectionMode = 0;		/* Mode for QRS detection using BP-ECG intervals (0: off, 1: on) */
    int qrsSeqBPECG[3000] = {0,};	/* array for locations of detected qrs complexes using BP-ECG intervals */
    int cntQRS = 0;			/* counter for number of detected qrs complexes using BP-ECG intervals */
    int qrsSeqWQRS[3000] = {0,};	/* array for QRS detection using WQRS algorithm */
    int cntQRSfromWQRS = 0;		/* counter for number of detected qrs complexes using WQRS algorithm */
    int endOfRecord = 150000;		/* Maximum samples from a record (we assumed that a record has 10 minutes samples = 250Hz * 600s) */



    /* First argument is name of this program */
    pname = prog_name(argv[0]);

    /* Process input arguments */
    for (i = 1; i < argc; i++) {
	if (*argv[i] == '-') switch (*(argv[i]+1)) {
	    case 'd':	/* dump filter data */
		dflag = 1;
		break;
	    case 'f':	/* starting time */
		if (++i >= argc) {
		    (void)fprintf(stderr, "%s: time must follow -f\n", pname);
		    exit(1);
		}
		from = i;
		break;
	    case 'h':	/* help requested */
		help();
		exit(0);
		break;
	    case 'H':	/* operate in WFDB_HIGHRES mode */
		gvmode = WFDB_GVPAD | WFDB_HIGHRES;
		break;
	    case 'j':	/* annotate J-points (ends of QRS complexes) */
		jflag = 1;
		break;
	    case 'm':	/* threshold */
		if (++i >= argc || (Tm = atoi(argv[i])) <= 0) {
		    (void)fprintf(stderr, "%s: threshold ( > 0) must follow -m\n",
			    pname);
		    exit(1);
		}
		break;
	    case 'p':	/* specify power line (mains) frequency */
		if (++i >= argc || (PWFreq = atoi(argv[i])) <= 0) {
		    (void)fprintf(stderr,
			    "%s: power line frequency ( > 0) must follow -p\n",
			    pname);
		    exit(1);
		}
		break;
	    case 'r':	/* record name */
		if (++i >= argc) {
		    (void)fprintf(stderr, "%s: input record name must follow -r\n",
			    pname);
		    exit(1);
		}
		record = argv[i];
		break;
	    case 'R':	/* resample */
		Rflag = 1;
		break;
	    case 's':	/* signal */
		if (++i >= argc) {
		    (void)fprintf(stderr,
			    "%s: signal number or name must follow -s\n",
			    pname);
		    exit(1);
		}
		sig = i;	/* remember the argument until the record is open */
		break;
	    case 't':	/* end time */
		if (++i >= argc) {
		    (void)fprintf(stderr, "%s: time must follow -t\n", pname);
		    exit(1);
		}
		to = i;
		break;
	    case 'v':	/* verbose mode */
		vflag = 1;
		break;
	    default:
		(void)fprintf(stderr, "%s: unrecognized option %s\n", pname,
			argv[i]);
		exit(1);
	}
	else {
	    (void)fprintf(stderr, "%s: unrecognized argument %s\n", pname,
		    argv[i]);
	    exit(1);
	}
    }
    if (record == NULL) {
	help();
	exit(1);
    }


    /* set up the mode for getting vector  */
    if (gvmode == 0 && (p = getenv("WFDBGVMODE")))
	gvmode = atoi(p);
    setgvmode(gvmode|WFDB_GVPAD);

    /* Check whether input record is valid or not */
    if ((nsig = isigopen(record, NULL, 0)) < 1) exit(2);
    if ((s = (WFDB_Siginfo *)malloc(nsig * sizeof(WFDB_Siginfo))) == NULL||(v=malloc(nsig * sizeof(WFDB_Sample)))==NULL) {
	(void)fprintf(stderr, "%s: insufficient memory\n", pname);
	exit(2);
    }

    /* Open input record into signal information 's' */
    if ((nsig = isigopen(record, s, nsig)) < 1) exit(2);
    sps = sampfreq((char *)NULL);
    if (sps < PWFreq) {
	(void)fprintf(stderr, "%s: sampling frequency (%g Hz) is too low%s",
		pname, sps,
		(gvmode & WFDB_HIGHRES) ? "\n" : ", try -H option\n");
	exit(3);
    }

    if (gvmode & WFDB_HIGHRES)
	setafreq(sampfreq((char *)NULL));

    /* Set output annotation's extension */
    a.name = "qrs"; a.stat = WFDB_WRITE;
    if (annopen(record, &a, 1) < 0) exit(2);
    if (sig >= 0) sig = findsig(argv[sig]);
    if (sig < 0 || sig >= nsig) sig = 0;
    if ((gain = s[sig].gain) == 0.0) gain = WFDB_DEFGAIN;
    if (Rflag) {
	if (PWFreq == 60.0) setifreq(sps = 250.);
	else setifreq(sps = 250.);
    }

    /* Set range with starting and end time */
    if (from > 0L) {
	if ((from = strtim(argv[from])) < 0L)
	    from = -from;
    }
    if (to > 0L) {
	if ((to = strtim(argv[to])) < 0L)
	    to = -to;
    } else {
	to = strtim("e");
    }

    /* Initialize annotation structure */
    annot.subtyp = annot.num = 0;
    annot.chan = sig;
    annot.aux = NULL;
    Tm = muvadu((unsigned)sig, Tm);
    samplingInterval = 1000.0/sps;

    /* Parameters for WQRS algorithm */
    lfsc = 1.25*gain*gain/sps;		/* length function scale constant */
    spm = 60 * sps;
    next_minute = from + spm;
    LPn = sps/PWFreq; 			/* The LP filter will have a notch at the power line (mains) frequency */
    if (LPn > 8)  LPn = 8;		/* avoid filtering too agressively */
    LP2n = 2 * LPn;
    EyeClosing = sps * EYE_CLS; 	/* set eye-closing period */
    ExpectPeriod = sps * NDP;		/* maximum expected RR interval */
    LTwindow = sps * MaxQRSw;   	/* length transform window size */

    /* Read samples of the currently open input signals */
    (void)sample(sig, 0L);

    /* dump mode */
    if (dflag) {
	for (t = from; t < to || (to == 0L && sample_valid()); t++)
	    printf("%6d\t%6d\n", sample(sig, t), ltsamp(t));
	exit(0);
    }

    /* verbose mode */
    if (vflag) {
	printf("\n------------------------------------------------------\n");
	printf("Record Name:             %s\n", record);
	printf("Total Signals:           %d  (", nsig);
	for (i = 0; i < nsig - 1; i++)
	    printf("%d, ", i);
	printf("%d)\n", nsig - 1);
	printf("Sampling Frequency:      %.1f Hz\n", sps);
	printf("Sampling Interval:       %.3f ms\n", samplingInterval);
	printf("Signal channel used for detection:    %d\n", sig);
	printf("Eye-closing period:      %d samples (%.0f ms)\n", EyeClosing, EyeClosing*samplingInterval);
	printf("Minimum threshold:       %d A/D units (%d microvolts)\n", Tm, adumuv(sig, Tm));
	printf("Power line frequency:    %d Hz\n", PWFreq);
	printf("\n------------------------------------------------------\n\n");
	printf("Processing:\n");
    }

    /* Average the first 8 seconds of the length-transformed samples
       to determine the initial thresholds Ta and T0. The number of samples
       in the average is limited to half of the ltsamp buffer if the sampling
       frequency exceeds about 2 KHz. */
    if ((t1 = strtim("8")) > BUFLN*0.9)
	t1 = BUFLN/2;
    t1 += from;
    for (T0 = 0, t = from; t < t1 && sample_valid(); t++)
	T0 += ltsamp(t);
    T0 /= t1 - from;
    Ta = 3 * T0;

    /* Main loop */
    for (t = from; t < to || (to == 0L && sample_valid()); t++) {
	static int learning = 1, T1;

	if (learning) {
	    if (t > t1) {
		learning = 0;
		T1 = T0;
		t = from;	/* start over */
	    }
	    else
		T1 = 2*T0;
	}

	/* Compare a length-transformed sample against T1. */
	if (ltsamp(t) > T1) {	/* found a possible QRS near t */
	    timer = 0; /* used for counting the time after previous QRS */
	    max = min = ltsamp(t);
	    for (tt = t+1; tt < t + EyeClosing/2; tt++)
		if (ltsamp(tt) > max) max = ltsamp(tt);
	    for (tt = t-1; tt > t - EyeClosing/2; tt--)
		if (ltsamp(tt) < min) min = ltsamp(tt);
	    if (max > min+10) { /* There is a QRS near tt */
		/* Find the QRS onset (PQ junction) */
		onset = max/100 + 2;
		tpq = t - 5;
		for (tt = t; tt > t - EyeClosing/2; tt--) {
		    if (ltsamp(tt)   - ltsamp(tt-1) < onset &&
			    ltsamp(tt-1) - ltsamp(tt-2) < onset &&
			    ltsamp(tt-2) - ltsamp(tt-3) < onset &&
			    ltsamp(tt-3) - ltsamp(tt-4) < onset) {
			tpq = tt - LP2n;	/* account for phase shift */
			break;
		    }
		}

		if (!learning) {
		    /* Check that we haven't reached the end of the record. */
		    (void)sample(sig, tpq);
		    if (sample_valid() == 0) break;
		    /* Record an annotation at the QRS onset */
		    qrsSeqWQRS[cntQRSfromWQRS++] = tpq;
		    /* Following section for record an annotation will be used at the end of this loop */
		    /* RECORDING START */
		    /*
		    annot.time = tpq;
		    annot.anntyp = NORMAL;
		    if (tpq > 0 ) {
			if (putann(0, &annot) < 0) {
			    wfdbquit();
			    exit(1);
			}
		    }*/
		    /* RECORDING END */
		    if (jflag) {
			/* Find the end of the QRS */
			for (tt = t, tj = t + 5; tt < t + EyeClosing/2; tt++) {
			    if (ltsamp(tt) > max - (max/10)) {
				tj = tt;
				break;
			    }
			}
			(void)sample(sig, tj);
			if (sample_valid() == 0) break;
			/* Record an annotation at the J-point */
			annot.time = tj;
			annot.anntyp = JPT;
			if(putann(0, &annot) < 0) {
			    wfdbquit();
			    exit(1);
			}
		    }
		}

		/* Adjust thresholds */
		Ta += (max - Ta)/10;
		T1 = Ta / 3;


		/* Lock out further detections during the eye-closing period */
		t += EyeClosing;
	    }
	}
	else if (!learning) {
	    /* Once past the learning period, decrease threshold if no QRS
	       was detected recently. */
	    if (++timer > ExpectPeriod && Ta > Tm) {
		Ta--;
		T1 = Ta / 3;
	    }      
	}

	/* Keep track of progress by printing a dot for each minute analyzed */
	if (t >= next_minute) {
	    next_minute += spm;
	    (void)fprintf(stderr, ".");
	    (void)fflush(stderr);
	    if (++minutes >= 60) {
		(void)fprintf(stderr, " %s\n", timstr(t));
		minutes = 0;
	    }
	}
    }
    if (minutes) (void)fprintf(stderr, " %s\n", timstr(t));

    (void)free(lbuf);
    (void)free(ebuf);

    /* =========================  */
    /* Section for BP-ECG method  */
    /* =========================  */

    /* Turn back to the start point for getvec() */
    reset_time = strtim("0");
    isigsettime(reset_time);

    /* Read the first 150ms of ECG and BP samples */
    for(i=0;i<150;i++) {
	(void)getvec(v);
	ecg_temp[i] = ecg_array[i]  = v[0];
	bp_temp[i]  = bp_array[i]   = v[1];
	current_time++;
    }
    
    /* Part 1. Estimate initial parameter for BP-ECG intervals */
    /*         until processing 50 maximum points of BP signal */ 
    do {
	/* Get maximum points in three sections of BP array */
	bp_left	    = max_point(bp_array, 0, 49);
	bp_mid	    = max_point(bp_array, 50, 99);
	bp_right    = max_point(bp_array, 100, 149);

	/* Check the middle point is the most biggest value ( protruding shape of BP signal: |_|/\|_| ) */ 
	if( (bp_array[bp_left] <= bp_array[bp_mid]) && (bp_array[bp_mid] >= bp_array[bp_right]) ) {
	    /* Check weather the detection mode is on or off  */
	    if(qrsDetectionMode == 0) {
		/* Detection Off => On */
		qrsDetectionMode = 1;
		cntBpMax++;

		/* 1. Find maximum point on ECG signal from the start point and index of maximum point on BP signal */
		if(current_time == 150) {
		    ecg_max = max_point(ecg_temp, 0, bp_mid);
		} else if(current_time == 200) {	
		    ecg_max = max_point(ecg_temp, 0, bp_mid+50);
		} else {
		    ecg_max = max_point(ecg_temp, bp_mid-100, bp_mid+100); // ??
		}

		/* 2. detect qrs complex with maximum value on ECG signal */
		qrs_time = base_time + ecg_max - offset_time;

		/* Save index of qrs complex and its interval */
		qrsSeqBPECG[cntQRS++] = qrs_time;
		qrs_interval += (bp_mid-ecg_max);

		/* 3. Searching window moves right with 200ms and fill with new samples */
		for(i = 0; i < 100; i++) {
		    ecg_array[i]    = ecg_array[i+50];
		    bp_array[i]	    = bp_array[i+50];
		}
		for(i = 0; i < 50; i++) {
		    (void)getvec(v);
		    ecg_t[i]	= ecg_array[i+100]  = v[0];
		    bp_t[i]	= bp_array[i+100]   = v[1];
		}

		if(current_time == 150) {
		    for(i = 0; i < 50; i++) { ecg_temp[i+150]	= ecg_t[i]; }
		} else if(current_time == 200){
		    for(i = 0; i < 50; i++) { ecg_temp[i+200]	= ecg_t[i]; }
		} else {
		    for(i = 0; i < 200; i++){ ecg_temp[i]	= ecg_temp[i+50]; }
		    for(i = 0; i < 50;  i++){ ecg_temp[i+100]	= ecg_t[i]; }
		}

		current_time += 50;
		base_time += 50;

	    } else { /* When detection mode is on */
		/* Turn off detection mode */
		qrsDetectionMode = 0;

		/* Searching window moves to the right with 200ms without detecting process */
		for(i = 0; i < 100; i++) {
		    ecg_array[i]    = ecg_array[i+50];
		    bp_array[i]	    = bp_array[i+50];
		}
		for(i = 0; i < 50; i++) {
		    (void)getvec(v);
		    ecg_t[i]	= ecg_array[i+100]  = v[0];
		    bp_t[i]	= bp_array[i+100]   = v[1];
		}

		if(current_time == 150) {
		    for(i = 0; i < 50; i++) { ecg_temp[i+150]	= ecg_t[i]; }
		} else if(current_time == 200) {
		    for(i = 0; i < 50; i++) { ecg_temp[i+200]	= ecg_t[i]; }
		} else {
		    for(i = 0; i < 200; i++){ ecg_temp[i]	= ecg_temp[i+50]; }
		    for(i = 0; i < 50;  i++){ ecg_temp[i+100]	= ecg_t[i]; }
		}	 

		current_time += 50;
		base_time += 50;
	    }
	} else {
	    /* When the shape of BP signal is not protruding */

	    /* Turn off detection mode */
	    qrsDetectionMode =0;

	    /* Searching window moves to the right with 200ms without detecting process */
	    for(i = 0; i < 100; i++) {
		ecg_array[i]	= ecg_array[i+50];
		bp_array[i]	= bp_array[i+50];
	    }
	    for(i = 0; i < 50; i++) {
		(void)getvec(v);
		ecg_t[i]    = ecg_array[i+100]	= v[0];
		bp_t[i]	    = bp_array[i+100]	= v[1];
	    }

	    for(i = 0; i < 200; i++) { ecg_temp[i]	= ecg_temp[i+50];  }
	    for(i = 0; i < 50;  i++) { ecg_temp[i+100]	= ecg_t[i]; }

	    current_time += 50;
	    base_time += 50;
	} 

    /* Until predefined observation */
    } while (cntBpMax == numObservation);
    
    /* Calculate the average of collected intervals */
    qrs_interval = qrs_interval / numObservation;

    /* Part 2. Find QRS complex using estimated intervals between BP-ECG maximum points */
    do {
	/* Get maximum points in three sections of BP array */
	bp_left	    = max_point(bp_array, 0, 49);
	bp_mid	    = max_point(bp_array, 50, 99);
	bp_right    = max_point(bp_array, 100, 149);

	/* Check the middle point is the most biggest value ( protruding shape of BP signal: |_|/\|_| ) */ 
	if( (bp_array[bp_left] <= bp_array[bp_mid]) && (bp_array[bp_mid] >= bp_array[bp_right]) ) { 
	    /* Check weather the detection mode is on or off  */
	    if(qrsDetectionMode == 0) {
		/* Detection Off => On */
		qrsDetectionMode = 1;

		/* Directly detect qrs complex with maximum value on BP signal and estimated QRS interval */
		qrs_time = base_time + bp_mid - qrs_interval - offset_time;

		/* Save index of qrs complex and its interval */
		qrsSeqBPECG[cntQRS++] = qrs_time;

		/* Searching window moves right with 200ms and fill with new samples */
		for(i=0;i<100;i++){
		    ecg_array[i]    = ecg_array[i+50];
		    bp_array[i]	    = bp_array[i+50];
		}
		for(i=0;i<50;i++){
		    (void)getvec(v);
		    ecg_t[i]	= ecg_array[i+100]  = v[0];
		    bp_t[i]	= bp_array[i+100]   = v[1];
		}

		for(i=0;i<200;i++){ ecg_temp[i] = ecg_temp[i+50];} 
		for(i=0;i<50;i++) { ecg_temp[i+100] = ecg_t[i]; }

		current_time += 50;
		base_time += 50;
	    } else { /* When detection mode is on */
		/* Turn off detection mode */
		qrsDetectionMode = 0;

		/* Searching window moves to the right with 200ms without detecting process */
		for(i=0;i<100;i++){
		    ecg_array[i]    = ecg_array[i+50];
		    bp_array[i]	    = bp_array[i+50];
		}
		for(i=0;i<50;i++){
		    (void)getvec(v);
		    ecg_t[i]	= ecg_array[i+100]  = v[0];
		    bp_t[i]	= bp_array[i+100]   = v[1];
		}

		for(i=0;i<200;i++){ ecg_temp[i]	    = ecg_temp[i+50]; }
		for(i=0;i<50;i++) { ecg_temp[i+100] = ecg_t[i]; }

		current_time += 50;
		base_time += 50;
	    }
	} else {
	    /* When the shape of BP signal is not protruding */
	    /* Turn off detection mode */
	    qrsDetectionMode = 0;

	    /* Searching window moves to the right with 200ms without detecting process */
	    for(i=0;i<100;i++){
		ecg_array[i] = ecg_array[i+50];
		bp_array[i] = bp_array[i+50];
	    }
	    for(i=0;i<50;i++){
		(void)getvec(v);
		ecg_t[i] = ecg_array[i+100] = v[0];
		bp_t[i] = bp_array[i+100] = v[1];
	    }

	    for(i=0;i<200;i++){ ecg_temp[i]	= ecg_temp[i+50]; }
	    for(i=0;i<50;i++) { ecg_temp[i+100]	= ecg_t[i]; }

	    current_time += 50;
	    base_time += 50;
	}
    }while(current_time != endOfRecord);



    //	printf("qrscomplex : %d  | %d\n",qrsc,ltqrsc);

    /* Merge the results of BP-ECG and WQRS */
    /*
    if(cntQRS<cntQRSfromWQRS){
	for(i=0;i<cntQRSfromWQRS;i++){
	    if (qrsSeqWQRS[i] > 0) {
		annot.time = qrsSeqWQRS[i]; annot.anntyp = NORMAL;
		(void)putann(0,&annot);
	    }
	} 
    }else{*/
	for(i=0;i<cntQRS;i++){
	    if (qrsSeqBPECG[i] > 0) {
		annot.time = qrsSeqBPECG[i]; annot.anntyp = NORMAL;
		(void)putann(0,&annot);
	    }
	}
    //}	
    //annot.time = qrs_time; annot.anntyp = NORMAL;

    //(void)putann(0,&annot);
    //for(i=0;i<qrsc;i++){
    //	printf("%d : %60d | %60d\n",i,qrs_seq[i],qrs_seq_lt[i]);
    //}


    wfdbquit();		        /* close WFDB files */
    fprintf(stderr, "\n");
    if (vflag) {
	printf("\n\nDone! \n\nResulting annotation file:  %s.qrs\n\n\n",
		record);
    }
    exit(0);
}

char *prog_name(s)
    char *s;
{
    char *p = s + strlen(s);

#ifdef MSDOS
    while (p >= s && *p != '\\' && *p != ':') {
	if (*p == '.')
	    *p = '\0';		/* strip off extension */
	if ('A' <= *p && *p <= 'Z')
	    *p += 'a' - 'A';	/* convert to lower case */
	p--;
    }
#else
    while (p >= s && *p != '/')
	p--;
#endif
    return (p+1);
}

static char *help_strings[] = {
    "usage: %s -r RECORD [OPTIONS ...]\n",
    "where RECORD is the name of the record to be analyzed, and OPTIONS may",
    "include any of:",
    " -d          dump unfiltered and filtered samples on standard output;",
    "             do not annotate",
    " -f TIME     begin at specified time (default: beginning of the record)",
    " -h          print this usage summary",
    " -H          read multifrequency signals in high resolution mode",
    " -j          find and annotate J-points (QRS ends) as well as QRS onsets",
    " -m THRESH   set detector threshold to THRESH (default: 100)", /* TmDEF */
    " -p FREQ     specify power line (mains) frequency (default: 60)",
    /* PWFreqDEF */
    " -R          resample input at 120 or 150 Hz, depending on power line",
    "             frequency (default: do not resample)",
    " -s SIGNAL   analyze specified signal (default: 0)",
    " -t TIME     stop at specified time (default: end of the record)",
    " -v          verbose mode",
    "If too many beats are missed, decrease THRESH;  if there are too many extra",
    "detections, increase THRESH.",
    NULL
};

void help()
{
    int i;

    (void)fprintf(stderr, help_strings[0], pname);
    for (i = 1; help_strings[i] != NULL; i++)
	(void)fprintf(stderr, "%s\n", help_strings[i]);
}

/* Return the location of maximum point in input array between start and end */
int max_point(int *array, int start, int end)
{
    int i = 0;
    int max = 0;
    int loc = 0;
    max = array[start];
    loc = start;
    for(i = start; i <= end; i++){
	if(max <= array[i]) {
	    max = array[i];
	    loc = i;
	} else {
	    continue;
	}
    }

    return loc; 
}
